/**
 * @file m13_ra6m3_ek_camera.c
 * @brief M13-RA6M3-EK OV7740
 *
 * @section License
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright (C) 2022 M13DESIGN SARL. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http:// www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author M13DESIGN SARL (www.m13design.fr)
 * @version 1.0.0
 **/

// Includes
#include "FreeRTOS.h"
#include "common_data.h"
#include "hal_data.h"
#include "r_transfer_api.h"
#include "r_dmac.h"
#include "m13_ra6m3_ek_camera.h"
#include "m13_ra6m3_ek_vcom.h"
#include "m13_ra6m3_ek_config.h"
#include "m13_ra6m3_ek_lcd.h"
#include "OV7670.h"



// Macro definitions
#define OV7740_CAM_PWR_DN   BSP_IO_PORT_05_PIN_08
#define OV7740_CAM_RST      BSP_IO_PORT_05_PIN_07

// Macro Temps
#define OV7670_PID_DEFAULT  0x76
#define OV7670_VER_DEFAULT  0x73


// Functions
static void  bsp_camera_delay_ms (uint32_t wait);

// External functions
extern void r_pdc_transfer_callback(pdc_instance_ctrl_t *p_ctrl);
void g_pdc_user_callback (pdc_callback_args_t * p_args);
void g_pdc_transfer_callback(dmac_callback_args_t *p_args);

// Global variables
static ioport_instance_ctrl_t camera_pin_ctrl;
bsp_i2c_handle_t camera_handle = BSP_I2C_INVALID_HANDLE;

uint8_t bsp_camera_buffer [ BSP_CAM_WIDTH * BSP_CAM_HEIGHT * BSP_CAM_BYTES_PER_PIXEL ] BSP_PLACE_IN_SECTION(BSP_CAMERA_RAM_SECTION) BSP_ALIGN_VARIABLE(8);

// Callback functions
void g_pdc_user_callback (pdc_callback_args_t * p_args)
{
    BaseType_t context_switch;

    if (PDC_EVENT_TRANSFER_COMPLETE == p_args->event)
    {
        xSemaphoreGiveFromISR(pdc_semaphore, &context_switch);
    }
    else
    {
       // xSemaphoreGiveFromISR(pdc_semaphore, &context_switch);
    }
}
void g_pdc_transfer_callback(dmac_callback_args_t *p_args)
{
    r_pdc_transfer_callback ((pdc_instance_ctrl_t*) p_args->p_context);
}


/*-------------------------DMAC configuration------------------------*/

dmac_instance_ctrl_t g_transfer0_ctrl;

transfer_info_t g_transfer0_info =
{
  .dest_addr_mode            = TRANSFER_ADDR_MODE_INCREMENTED,
  .repeat_area               = TRANSFER_REPEAT_AREA_SOURCE,
  .irq                       = TRANSFER_IRQ_END,
  .chain_mode                = TRANSFER_CHAIN_MODE_DISABLED,
  .src_addr_mode             = TRANSFER_ADDR_MODE_FIXED,
  .size                      = TRANSFER_SIZE_4_BYTE,
  .mode                      = TRANSFER_MODE_BLOCK,
  .p_dest                    = (void*) NULL,
  .p_src                     = (void const*) NULL,
  .num_blocks                = 1,
  .length                    = 8,
};

const dmac_extended_cfg_t g_transfer0_extend =
{
  .offset                    = 1,
#if defined(VECTOR_NUMBER_DMAC0_INT)
  .irq                       = VECTOR_NUMBER_DMAC0_INT,
#else
  .irq                       = FSP_INVALID_VECTOR,
#endif
  .ipl                       = (8),
  .channel                   = 0,
  .p_callback                = g_pdc_transfer_callback,
  .p_context                 = &g_pdc0_ctrl,
  .activation_source         = ELC_EVENT_PDC_RECEIVE_DATA_READY,
};

const transfer_cfg_t g_transfer0_cfg =
{ .p_info = &g_transfer0_info, .p_extend = &g_transfer0_extend, };

/* Instance structure to use this module. */
const transfer_instance_t g_transfer0 =
{ .p_ctrl = &g_transfer0_ctrl, .p_cfg = &g_transfer0_cfg, .p_api = &g_transfer_on_dmac };
#define RA_NOT_DEFINED (UINT32_MAX)

/*---------------------END OF DMAC configuration---------------------*/

// Camera configuration
#undef RA_NOT_DEFINED

pdc_instance_ctrl_t g_pdc0_ctrl;

pdc_cfg_t g_pdc0_cfg =
{
  .bytes_per_pixel       = BSP_CAM_BYTES_PER_PIXEL,
  .clock_division        = PDC_CLOCK_DIVISION_2,  // Not used as the camera clock is provided by an on board 27MHz oscillator
  .endian                = PDC_ENDIAN_LITTLE,
  .hsync_polarity        = PDC_HSYNC_POLARITY_HIGH,
  .vsync_polarity        = PDC_VSYNC_POLARITY_LOW,
  .p_buffer              = (uint8_t*) NULL,
  .p_callback            = g_pdc_user_callback,
  /** If NULL then do not add & */
#if defined(NULL)        
  .p_context             = NULL,
#else                    
  .p_context             = &NULL,
#endif                   
  .p_extend              = NULL,
  .p_lower_lvl_transfer  = &g_transfer0,
  .x_capture_pixels      = BSP_CAM_WIDTH,
  .y_capture_pixels      = BSP_CAM_HEIGHT,
  .x_capture_start_pixel = 0,
  .y_capture_start_pixel = 0,
  .pdc_ipl               = (2),
  .transfer_req_ipl      = (BSP_IRQ_DISABLED),
#if defined(VECTOR_NUMBER_PDC_INT)
  .pdc_irq               = VECTOR_NUMBER_PDC_INT,
#else
  .pdc_irq               = FSP_INVALID_VECTOR,
#endif

#if defined(VECTOR_NUMBER_PDC_RECEIVE_DATA_READY)
  .transfer_req_irq      = VECTOR_NUMBER_PDC_RECEIVE_DATA_READY,
#else
  .transfer_req_irq      = FSP_INVALID_VECTOR,
#endif
};

const pdc_instance_t g_pdc0 =
{ .p_ctrl = &g_pdc0_ctrl, .p_cfg = &g_pdc0_cfg, .p_api = &g_pdc_on_pdc, };

const rm_freertos_port_parameters_t camera_thread_parameters =
{ .p_context = (void*) NULL, };

/*---------------------END OF CAMERA configuration---------------------*/

/**
 * @name       bsp_lcd_io_init
 * @brief      Initialize / Configure LCD pins
 * @return     None
 */
void bsp_camera_io_init (void)
{
    ioport_cfg_t camera_pin_cfg;

    camera_pin_cfg.number_of_pins = sizeof(m13_ra6m3_ek_camera_pin_data)/sizeof(ioport_pin_cfg_t);

    camera_pin_cfg.p_pin_cfg_data = &m13_ra6m3_ek_camera_pin_data[0];

    R_IOPORT_Open(&camera_pin_ctrl, &camera_pin_cfg);
}


/**
 * @name       bsp_camera_init
 * @brief      OV7740 initialization
 * @return     Error code (see #ErrorCode enumeration)
 **/
bsp_status_t bsp_camera_init(void)
{
    bsp_status_t error;
    fsp_err_t   err;
    uint8_t value;

    // Configure camera pins
    bsp_camera_io_init();

    // Power off camera module
    bsp_camera_power( POWER_DOWN );

    // 10ms delay
    bsp_camera_delay_ms(10);

	// Power on camera module
    bsp_camera_power( POWER_UP );

	// Reset Camera module
    bsp_camera_reset();

    // 10ms delay
    bsp_camera_delay_ms(10);

    // Open camera module
    err = R_PDC_Open(&g_pdc0_ctrl, &g_pdc0_cfg);
    if (FSP_SUCCESS != err)
    {
       return BSP_STATUS_FAILURE;
    }

    // Retrieve I2C channel 0 handle
    camera_handle = bsp_i2c_open(BSP_I2C_CHANNEL_0);

    if (camera_handle == BSP_I2C_INVALID_HANDLE)
    {
        return BSP_STATUS_INVALID_CHANNEL;
    }

#if BSP_CONFIG_CAMERA_I2C_TEST

    bsp_camera_dump_register();

#endif

    // Read OV7740_PIDH register
    error = bsp_camera_read_reg(OV7740_PIDH, &value);
    if(error)
    {
        return error;
    }

    // Verify device High identifier
    if( (value != OV7740_PIDH_DEFAULT) & (value != OV7670_PID_DEFAULT))
    {
        PRINTF_INFO(" Camera PIDH read = 0x%02X \r\n", value);
        return BSP_STATUS_INVALID_VERSION;
    }

    // Read OV7740_PIDL register
    error = bsp_camera_read_reg(OV7740_PIDL, &value);
    if(error)
    {
        return error;
    }

    // Verify device Low identifier
    if( (value != OV7740_PIDL_DEFAULT) & (value != OV7670_VER_DEFAULT))
    {
        PRINTF_INFO(" Camera PIDL read = 0x%02X \r\n", value);
        return BSP_STATUS_INVALID_VERSION;
    }

#if BSP_CONFIG_CAMERA_IS_OV7740
    // Write camera config
    error = bsp_camera_write_array(OV7740_REG_TABLE);
    if (BSP_STATUS_OK != error)
    {
   	 return error;
    }
#endif

#if BSP_CONFIG_CAMERA_IS_OV7670
    // Write camera config
    error = bsp_camera_write_array(OV7670_REG_TABLE);
    if (BSP_STATUS_OK != error)
    {
     return error;
    }
#endif

#if BSP_CONFIG_CAMERA_I2C_TEST

    bsp_camera_dump_register();

#endif

    // Successful initialization
    PRINTF_INFO(" camera Init Success!\r\n");

    return BSP_STATUS_OK;
}

/**
 * @name       bsp_camera_write_reg
 * @brief      Write OV7740 register
 * @param[in]  address : Register address
 * @param[in]  data : data to be written
 * @return     Error code (see #ErrorCode enumeration)
 **/
bsp_status_t bsp_camera_write_reg(uint8_t address, uint8_t data)
{
    bsp_status_t error;

    // Write i2c
    error = bsp_i2c_write (camera_handle, BSP_I2C_SLAVE_ADDR_CAMERA, address, 1, &data, 1 );

    // Return status code
    return (error);
}

bsp_status_t bsp_camera_write_array(ov_reg_t *array)
{
	bsp_status_t error;

    while (0xFF != array->reg)
    //while (0xC9 != array->reg)
    {
    	error = bsp_camera_write_reg(array->reg, array->val);

        if (BSP_STATUS_OK != error)
        {
            return error;
        }

        array++;
     }
    // Return status code
    return (error);
}

/**
 * @name       bsp_camera_read_reg
 * @brief      Read OV7740 register
 * @param[in]  address : Register address
 * @param[out] data : data to be read
 * @return     Error code (see #ErrorCode enumeration)
 **/
bsp_status_t bsp_camera_read_reg(uint8_t address, uint8_t *data)
{
    bsp_status_t error;

    // Read i2c
    error = bsp_i2c_read (camera_handle, BSP_I2C_SLAVE_ADDR_CAMERA, address, 1, data, 1 );

    // Return status code
    return (error);
}

/**
 * @name       bsp_camera_read_reg
 * @brief      Read OV7740 registers
 * @param[in]  address : Register address
 * @param[in]  data_length : data table length
 * @param[out] data : data table to be read
 * @return     Error code (see #ErrorCode enumeration)
 **/
bsp_status_t bsp_camera_read_regs(uint8_t address, uint8_t *data, size_t data_length)
{
    bsp_status_t error;

    // Read i2c
    error = bsp_i2c_read (camera_handle, BSP_I2C_SLAVE_ADDR_CAMERA, address, 1, data, data_length);

    // Return status code
    return (error);
}

/**
 * @name       bsp_camera_reset
 * @brief      Reset OV7740 device
 * @param[in]  None
 * @param[out] None
 * @return     BSP_STATUS_OK
 **/
bsp_status_t bsp_camera_reset(void)
{
	bsp_status_t error = BSP_STATUS_OK;
	
    //vTaskDelay (10 / portTICK_PERIOD_MS);

	R_BSP_PinAccessEnable();

#if BSP_CONFIG_ENABLE_TEST

	/* Active high reset behavior */

    R_BSP_PinWrite(OV7740_CAM_RST, BSP_IO_LEVEL_HIGH);

    // 10ms delay
    bsp_camera_delay_ms(10);
	
	R_BSP_PinWrite(OV7740_CAM_RST, BSP_IO_LEVEL_LOW);


#else

	/* Active low reset behavior */

    R_BSP_PinWrite(OV7740_CAM_RST, BSP_IO_LEVEL_LOW);

    // 10ms delay
    bsp_camera_delay_ms(10);

    R_BSP_PinWrite(OV7740_CAM_RST, BSP_IO_LEVEL_HIGH);
#endif

    R_BSP_PinAccessDisable();

	return (error);
}

/**
 * @name       bsp_camera_power
 * @brief      Turn on or off OV7740 camera
 * @param[in]  Pin state
 * @param[out] None
 * @return     BSP_STATUS_OK
 **/
bsp_status_t bsp_camera_power(bsp_camera_power_t power_up_or_down)
{
	bsp_status_t error = BSP_STATUS_OK;
	
	R_BSP_PinAccessEnable();

    R_BSP_PinWrite(OV7740_CAM_PWR_DN, (bsp_io_level_t)power_up_or_down);
	
	R_BSP_PinAccessDisable();
	
	return (error);
}

/**
 * @name       bsp_camera_capture_image
 * @brief      Launch an image capture
 * @param[in]  None
 * @param[out] None
 * @return     Error code (see #ErrorCode enumeration)
 **/
bsp_status_t bsp_camera_capture_image(void)
{
	fsp_err_t   error;
	BaseType_t  status;

	error = R_PDC_CaptureStart(&g_pdc0_ctrl, bsp_camera_buffer);

	if (FSP_SUCCESS != error)
    {
		return BSP_STATUS_FAILURE;
    }

	// Wait for capture to finish
	status = xSemaphoreTake(pdc_semaphore, ( TickType_t ) portMAX_DELAY );

	return BSP_STATUS_OK;
}

#define RANGE_LIMIT(x)        ( x > 255 ? 255 : (x < 0 ? 0 : x) )
/**
 * @name       bsp_camera_convert_display
 * @brief      Convert an image from yuv422 to rgb565 and send it to a display buffer and centers the image
 * @param[in]  inbuf : Pointer to camera buffer in YUV422
 * @param[in]  outbuf : Pointer to display buffer
 * @param[in]  width : Camera width
 * @param[in]  height : Camera height
 * @return     Error code (see #ErrorCode enumeration)
 **/
bsp_status_t bsp_camera_convert_display(const void* inbuf, void* outbuf, uint32_t width, uint32_t height)
{
	uint32_t rows, columns;
	int32_t  y, u, v;
	uint32_t r, g, b;
	uint8_t  *yuv_buf;
	uint16_t *display_buffer;
	uint32_t y_pos,u_pos,v_pos;

	int32_t temp;
	uint32_t pixel_data;
	uint32_t x_start, y_start;

	yuv_buf = (uint8_t *)inbuf;
	display_buffer = (uint16_t *)outbuf;


	// Initialize Y U V position
	y_pos = 1; //1
	u_pos = 0; //0
	v_pos = 2; //2

	// Calculate image starting centered position on the lcd
	x_start = (uint32_t) ( LCD_WIDTH - width ) / 2;
	y_start = (uint32_t) ( LCD_HEIGHT - height ) / 2;

	for (rows = 0; rows < height; rows++)
	{
		for (columns = 0; columns < width; columns++)
		{
			// Extract pixel Y U V byte from buffer

		    y = yuv_buf[y_pos];
			u = yuv_buf[u_pos] - 128;
			v = yuv_buf[v_pos] - 128;

			// Convert YUV422 to RGB888
			temp = y + v + ( (v * 103) >> 8 );
			r = RANGE_LIMIT( temp );

			temp = y - ( (u * 88) >> 8 ) - ( (v * 183) >> 8 );
			g = RANGE_LIMIT( temp );

			temp = y + u + ( (u * 198) >> 8 );
			b = RANGE_LIMIT( temp );

			// Convert RGB888 to RGB565
			pixel_data = ( ( (r & 0xF8) << 8 ) | ( (g & 0xFC) << 3 ) | ( (b & 0xF8) >> 3 ) );

			// Send pixel to display buffer and center the image
			display_buffer[ ( ( rows + y_start ) * LCD_WIDTH ) - (columns + x_start) ] = (uint16_t) pixel_data;

			// Move to next pixel
			y_pos += 2;

			// Move to next set of UV
			if (columns & 0x01)
			{
				u_pos += 4;
				v_pos += 4;
			}
		}
	}
	
	return BSP_STATUS_OK;
}

#define REG_TABLE_SIZE   0xFF

bsp_status_t bsp_camera_dump_register (void)
{
    bsp_status_t error;
    uint8_t dump_data[0xFF];

    PRINTF_INFO(" Start of Dump \r\n");
    PRINTF_INFO("\r\n");

    for(uint16_t i =0; i<= 0xFF; i++)
    {
        dump_data[i] = 0xFF;
    }

    // Dump registers
    error = bsp_camera_read_regs(0x00, dump_data, REG_TABLE_SIZE);

    // Display dumped registers on Vcom
    for(uint16_t i =0; i<= 0xFF; i++)
    {
        PRINTF_INFO(" reg %02X = 0x%02X \r\n", i, dump_data[i]);
    }

    PRINTF_INFO(" End of Dump \r\n");
    PRINTF_INFO("\r\n");

    return error;
}
/**
 * @name       bsp_camera_delay_ms
 * @brief      delay function - Can be replaced by OS delay function
 * @param[in]  wait : delay in millisecond
 * @return     None
 **/
static void bsp_camera_delay_ms (uint32_t wait)
{
    uint32_t i;

    for (i = WAIT_TICK; ( (i >= 0) && (0 != wait) ); i--)
    {
        if (0 == i)
        {
            wait--;
            i = WAIT_TICK;
        }
    }
}
