/**
 * @file m13_ra6m3_ek_camera.h
 * @brief M13-RA6M3-EK OV7740
 *
 * @section License
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright (C) 2022 M13DESIGN SARL. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author M13DESIGN SARL (www.m13design.fr)
 * @version 1.0.0
 **/

#ifndef _M13_RA6M3_EK_CAM_H
#define _M13_RA6M3_EK_CAM_H

// Includes
#include "m13_ra6m3_ek_config.h"
#include "m13_ra6m3_ek_status.h"
#include "m13_ra6m3_ek_pinout.h"
#include "m13_ra6m3_ek_i2c.h"
#include "OV7670.h"
#include "r_pdc.h"
#include "r_pdc_api.h"

// OV7740 register structure
typedef struct _ov_reg {
    uint8_t reg;
    uint8_t val;
} ov_reg_t;

typedef enum
{
	POWER_UP      = BSP_IO_LEVEL_LOW,
	POWER_DOWN    = BSP_IO_LEVEL_HIGH,
} bsp_camera_power_t;

extern pdc_instance_ctrl_t g_pdc0_ctrl;

typedef enum
{
    BSP_CAM_VGA_WIDTH	       = 640,
    BSP_CAM_VGA_HEIGHT	       = 480,
    BSP_CAM_CIF_WIDTH	       = 352,
    BSP_CAM_CIF_HEIGHT	       = 288,
    BSP_CAM_QVGA_WIDTH	       = 320,
    BSP_CAM_QVGA_HEIGHT	       = 240,
    BSP_CAM_QCIF_WIDTH	       = 176,
    BSP_CAM_QCIF_HEIGHT	       = 144,
    BSP_CAM_QQVGA_WIDTH	       = 160,
    BSP_CAM_QQVGA_HEIGHT	   = 120,
} bsp_camera_size_list_t;

// Macro definitions
#define      BSP_CAM_BYTES_PER_PIXEL      (2U)
#define      WAIT_TICK                    (100000)

// Registers used for OV7740 & OV7670
#define OV7740_PIDH                       (0x0A)
#define OV7740_PIDH_DEFAULT               (0x77)
#define OV7740_PIDL                       (0x0B)
#define OV7740_PIDL_DEFAULT               (0x42)

// OV7740 specific configuration
#define OV7740_REG0C_MAX_EXPOSURE_Pos     (1)
#define OV7740_REG0C_MAX_EXPOSURE_Msk     (0x3u << OV7740_REG0C_MAX_EXPOSURE_Pos)
#define OV7740_REG0C_MAX_EXPOSURE(value)  ((OV7740_REG0C_MAX_EXPOSURE_Msk & ((value) << OV7740_REG0C_MAX_EXPOSURE_Pos)))
#define OV7740_REG0C_MIRROR_ENABLE        (0x1u << 6)
#define OV7740_REG0C_FLIP_ENABLE          (0x1u << 7)
#define OV7740_REG0C_YUV_SWAP_ENABLE      (0x1u << 4)
#define OV7740_REG0C_YUV_BIT_SWAP         (0x1u << 3)

/********** Camera is OV7740 **********/
#if (BSP_CONFIG_CAMERA_IS_OV7740 == 1)

// Configuration table for OV7740 camera in YUV QVGA
#if ( (BSP_CONFIG_CAMERA_YUV == 1) & (BSP_CONFIG_CAMERA_QVGA == 1) )
extern const ov_reg_t                 OV7740_YUV422_QVGA[];
#define      OV7740_REG_TABLE         (ov_reg_t *)OV7740_YUV422_QVGA

typedef enum
{
    BSP_CAM_WIDTH              = BSP_CAM_QVGA_WIDTH,
    BSP_CAM_HEIGHT             = BSP_CAM_QVGA_HEIGHT,
    BSP_CAM_RESOLUTION         = BSP_CAM_WIDTH * BSP_CAM_HEIGHT * BSP_CAM_BYTES_PER_PIXEL,
} bsp_camera_current_size_t;
#endif

#endif
/********** End of Configuration ******/

/********** Camera is OV7670 **********/
#if (BSP_CONFIG_CAMERA_IS_OV7670 == 1)

// Configuration table for OV7670 camera
extern const ov_reg_t                 OV7670_config[];
#define      OV7670_REG_TABLE         (ov_reg_t*) OV7670_config

// Camera is in RGB format
#if BSP_CONFIG_CAMERA_RGB

// RGB565 register configurations
#define OV7670_COM7_CONFIG    ( OV7670_COM7_RGB )
#define OV7670_COM15_CONFIG   ( OV7670_COM15_RGB565 | OV7670_COM15_00_FF )
#define OV7670_MTX1_CONFIG    ( 0xB3 )
#define OV7670_MTX2_CONFIG    ( 0xB3 )
#define OV7670_MTX3_CONFIG    ( 0x00 )
#define OV7670_MTX4_CONFIG    ( 0x3D )
#define OV7670_MTX5_CONFIG    ( 0xA7 )
#define OV7670_MTX6_CONFIG    ( 0xE4 )

// Custom resolution (Horizontal and Vertical capture resolutions must be filled in OV7670.h by default set to 272x480 to fit the board lcd after a 90° rotation)
#if (BSP_CONFIG_CAMERA_CUSTOM_RES)
typedef enum
{
    BSP_CAM_WIDTH              = OV_7670_CAPTURE_H_RES,
    BSP_CAM_HEIGHT             = OV_7670_CAPTURE_V_RES,
    BSP_CAM_RESOLUTION         = BSP_CAM_WIDTH * BSP_CAM_HEIGHT * BSP_CAM_BYTES_PER_PIXEL,
} bsp_camera_current_size_t;
#endif

// Camera resolution in QVGA which fits on the board lcd after a 90° rotation
#if (BSP_CONFIG_CAMERA_QVGA)
typedef enum
{
    BSP_CAM_WIDTH              = BSP_CAM_QVGA_HEIGHT,
    BSP_CAM_HEIGHT             = BSP_CAM_QVGA_WIDTH,
    BSP_CAM_RESOLUTION         = BSP_CAM_WIDTH * BSP_CAM_HEIGHT * BSP_CAM_BYTES_PER_PIXEL,
} bsp_camera_current_size_t;
#endif

// Camera resolution in QVGA which fits on the board lcd after a 90° rotation
#if (BSP_CONFIG_CAMERA_QQVGA)
typedef enum
{
    BSP_CAM_WIDTH              = BSP_CAM_QQVGA_HEIGHT,
    BSP_CAM_HEIGHT             = BSP_CAM_QQVGA_WIDTH,
    BSP_CAM_RESOLUTION         = BSP_CAM_WIDTH * BSP_CAM_HEIGHT * BSP_CAM_BYTES_PER_PIXEL,
} bsp_camera_current_size_t;
#endif

#else

// Camera is in YUV format
#define OV7670_COM7_CONFIG    ( OV7670_COM7_YUV )
#define OV7670_COM15_CONFIG   ( OV7670_COM15_00_FF )
#define OV7670_MTX1_CONFIG    ( 0x80 )
#define OV7670_MTX2_CONFIG    ( 0x80 )
#define OV7670_MTX3_CONFIG    ( 0x00 )
#define OV7670_MTX4_CONFIG    ( 0x22 )
#define OV7670_MTX5_CONFIG    ( 0x5E )
#define OV7670_MTX6_CONFIG    ( 0x80 )

// Set camera resolution and buffer size. Resolution by default is in QVGA(320x240). Fits on the board lcd without rotation
// Other resolutions not tested in YUV
typedef enum
{
    BSP_CAM_WIDTH              = BSP_CAM_QVGA_WIDTH,
    BSP_CAM_HEIGHT             = BSP_CAM_QVGA_HEIGHT,
    BSP_CAM_RESOLUTION         = BSP_CAM_WIDTH * BSP_CAM_HEIGHT * BSP_CAM_BYTES_PER_PIXEL,
} bsp_camera_current_size_t;

#endif

#endif
/********** End of Configuration ******/

//Camera related functions
void         bsp_camera_io_init           (void);
bsp_status_t bsp_camera_init              (void);


bsp_status_t bsp_camera_write_reg         (uint8_t reg_address, uint8_t data);
bsp_status_t bsp_camera_write_array       (ov_reg_t *array);

bsp_status_t bsp_camera_read_reg          (uint8_t address, uint8_t *data);
bsp_status_t bsp_camera_read_regs         (uint8_t address, uint8_t *data, size_t data_length);

bsp_status_t bsp_camera_reset             (void);
bsp_status_t bsp_camera_power             (bsp_camera_power_t power_up_or_down);
bsp_status_t bsp_camera_capture_image     (void);
bsp_status_t bsp_camera_convert_display   (const void* inbuf, void* outbuf, uint32_t width, uint32_t height);
bsp_status_t bsp_camera_dump_register     (void);

#endif
