/**
 * @file m13_ra6m3_ek_vcom.c
 * @brief M13-RA6M3-EK Virtual COM definition
 *
 * @section License
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright (C) 2020-2021 M13DESIGN SARL. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author M13DESIGN SARL (www.m13design.fr)
 * @version 1.0.0
 **/

#include "r_sci_uart.h"
#include "m13_ra6m3_ek_vcom.h"
#include "m13_ra6m3_ek_status.h"
#include <stdarg.h>

//VCOM IOPORT structure
static ioport_instance_ctrl_t vcom_pin_ctrl;

// Macro definitions
#define UART_CHANNEL         R_SCI4
#define UART_MODULE_STOP     UART4_MODULE_STOP

#define UART3_MODULE_STOP    R_MSTP_MSTPCRB_MSTPB28_Msk
#define UART4_MODULE_STOP    R_MSTP_MSTPCRB_MSTPB27_Msk


int vfprintf_custom(void *out, const char *format, va_list args);

// UART control
sci_uart_instance_ctrl_t vcom_ctrl;

baud_setting_t vcom_baud_setting =
{
     .abcse = 0,
     .abcs  = 0,
     .bgdm  = 1,
     .cks   = 0,
     .brr   = 64,
     .mddr  = (uint8_t) 256,
     .brme  = false
};

// UART extended configuration for UARTonSCI HAL driver
const sci_uart_extended_cfg_t vcom_cfg_extend =
{
     .clock             = SCI_UART_CLOCK_INT,
     .rx_edge_start     = SCI_UART_START_BIT_FALLING_EDGE,
     .noise_cancel      = SCI_UART_NOISE_CANCELLATION_DISABLE,
     .rx_fifo_trigger   = SCI_UART_RX_FIFO_TRIGGER_MAX,
     .p_baud_setting    = &vcom_baud_setting,

#if ( FSP_VERSION_MAJOR <= 2 )
     .uart_mode         = UART_MODE_RS232,
     .ctsrts_en         = SCI_UART_CTSRTS_RTS_OUTPUT,
     .flow_control_pin  = (bsp_io_port_pin_t) (0xFFFFU),
#else
     .flow_control    = SCI_UART_FLOW_CONTROL_RTS,
     .flow_control_pin  = (bsp_io_port_pin_t) UINT16_MAX,
#endif
};

// UART interface configuration
uart_cfg_t vcom_cfg =
{
     .channel       = 0,
     .data_bits     = UART_DATA_BITS_8,
     .parity        = UART_PARITY_OFF,
     .stop_bits     = UART_STOP_BITS_1,
     .p_callback    = NULL,
     .p_context     = NULL,
     .p_extend      = &vcom_cfg_extend,
     .p_transfer_tx = NULL,
     .p_transfer_rx = NULL,
     .rxi_ipl       = (12),
     .txi_ipl       = (12),
     .tei_ipl       = (12),
     .eri_ipl       = (12),
     .rxi_irq       = FSP_INVALID_VECTOR,
     .txi_irq       = FSP_INVALID_VECTOR,
     .tei_irq       = FSP_INVALID_VECTOR,
     .eri_irq       = FSP_INVALID_VECTOR,
};

// UART interface module
const uart_instance_t g_vcom =
{
     .p_ctrl = &vcom_ctrl,
     .p_cfg = &vcom_cfg,
     .p_api = &g_uart_on_sci
};

/**
 * @brief      Configure M13-RZA2M-EK virtual COM pins
 * @param[in]  None
 * @return     None
 */
void bsp_vcom_init(void)
{
    ioport_cfg_t vcom_pin_cfg;

    vcom_pin_cfg.number_of_pins = sizeof(m13_ra6m3_ek_vcom_pin_data)/sizeof(ioport_pin_cfg_t);
    vcom_pin_cfg.p_pin_cfg_data = &m13_ra6m3_ek_vcom_pin_data[0];

    //Configure vcom pins
    R_IOPORT_Open(&vcom_pin_ctrl, &vcom_pin_cfg);
}

/**
 * @brief      Open and configure the uart channel
 * @param[in]  channel : The desired channel to be configurated
 * @return     bsp_status_t : Error code (see #ErrorCode enumeration in )
 */
bsp_status_t bsp_vcom_open(bsp_vcom_channel_t channel)
{
    fsp_err_t error;

    vcom_cfg.channel = channel;

    error = R_SCI_UART_Open(&vcom_ctrl, &vcom_cfg);


    if(error == FSP_SUCCESS)
    {
        return BSP_STATUS_OPEN_FAILURE;
    }
    else
    {
        return BSP_STATUS_OK;
    }
}



/**
 * @brief Debug UART initialization
 * @param[in] baudrate UART baudrate
 **/

void m13_manual_debug_Init(uint32_t baudrate)
{
   //Disable protection
   R_SYSTEM->PRCR = 0xA50B;

   //Cancel  module stop state
   R_MSTP->MSTPCRB &= ~UART_MODULE_STOP;

   //Enable protection
   R_SYSTEM->PRCR = 0xA500;

#ifdef _NOT_RA8_
   //Disable SCI9 module /SCI4
   UART_CHANNEL->SCR = 0x00;

   //Unlock PFS registers
   R_PMISC->PWPR &= ~R_PMISC_PWPR_B0WI_Msk;
   R_PMISC->PWPR |= R_PMISC_PWPR_PFSWE_Msk;

   //Configure TXD9 (P6_2) // TXD4 P205
   R_PFS->PORT[2].PIN[5].PmnPFS = (4 << R_PFS_PORT_PIN_PmnPFS_PSEL_Pos) |
      R_PFS_PORT_PIN_PmnPFS_PMR_Msk | R_PFS_PORT_PIN_PmnPFS_PCR_Msk;

   //Configure RXD9 (P6_1) // RXD4 P206
   R_PFS->PORT[2].PIN[6].PmnPFS = (4 << R_PFS_PORT_PIN_PmnPFS_PSEL_Pos) |
      R_PFS_PORT_PIN_PmnPFS_PMR_Msk | R_PFS_PORT_PIN_PmnPFS_PCR_Msk;

   //Lock PFS registers
   R_PMISC->PWPR &= ~R_PMISC_PWPR_PFSWE_Msk;
   R_PMISC->PWPR |= R_PMISC_PWPR_B0WI_Msk;

   //Configure UART (8 bits, no parity, 1 stop bit) /SCI4
   UART_CHANNEL->SMR = 0;
   //Configure prescaler for baud rate generator
   UART_CHANNEL->SEMR = R_SCI0_SEMR_BGDM_Msk | R_SCI0_SEMR_ABCS_Msk;
   //Set baudrate
   UART_CHANNEL->BRR = (SystemCoreClock / 1) / (8 * baudrate) - 1;

   //Enable transmission and reception
   UART_CHANNEL->SCR |= R_SCI0_SCR_TE_Msk | R_SCI0_SCR_RE_Msk;

#else

#endif
}


/**
 * @brief Display the contents of an array
 * @param[in] stream Pointer to a FILE object that identifies an output stream
 * @param[in] prepend String to prepend to the left of each line
 * @param[in] data Pointer to the data array
 * @param[in] length Number of bytes to display
 **/

void m13_display_Array(FILE *stream,
   const char_t *prepend, const void *data, size_t length)
{
   uint_t i;

   for(i = 0; i < length; i++)
   {
      // Beginning of a new line?
      if((i % 16) == 0)
          m13_printf_custom(stream, "%s", prepend);

      // Display current data byte
      m13_printf_custom(stream, "%02X" " ", *((uint8_t *) data + i));

      // End of current line?
      if((i % 16) == 15 || i == (length - 1))
          m13_printf_custom(stream, "\r\n");
   }
}

/**
 * @brief Write character to stream
 * @param[in] c The character to be written
 * @param[in] stream Pointer to a FILE object that identifies an output stream
 * @return On success, the character written is returned. If a writing
 *   error occurs, EOF is returned
 **/

int_t m13_put_char(int_t c, FILE *stream)
{
   //Standard output or error output?
   if(stream == stdout || stream == stderr)
   {

      //Wait for the transmitter to be ready
      while((UART_CHANNEL->SSR & R_SCI0_SSR_TEND_Msk) == 0)
      {
      }

      //Send character
      UART_CHANNEL->TDR = (uint32_t) c;

      //Wait for the transfer to complete
      while((UART_CHANNEL->SSR & R_SCI0_SSR_TEND_Msk) == 0)
      {
      }

      //On success, the character written is returned
      return c;
   }
   //Unknown output?
   else
   {
      //If a writing error occurs, EOF is returned
      return EOF;
   }
}
